/*
 * General purpuse IEEE character device driver
 * for NEC 7210 based cards (and compatible).
 * Tested with 7210 and NI PC-IIa
 *
 * Written & (C) 1995,96 by Michael Boegl und Thomas Hohenberger
 * Lehrstuhl Experimentalphysik V
 * Universitaet Bayreuth
 * e-mail: ieee@lowt.phy.uni-bayreuth.de
 *
 *
 * Linux 2.2 support and use of interrupts by Andreas Wrl 1998-2001
 * Linux 2.4 support by Andreas Wrl 2002 
 * e-mail: Andreas(dot)Wuerl(at)gmx(dot)net
 * 
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation with the modifications stated in the 
 *   'COPYING' licence file.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   (See file 'GPL' distributed with this package.)
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Create each GPIB device file with:
 * mknod /dev/gpib/<name> c 30 <addr>
 * ex: mknod /dev/gpib/hp33120 c 30 10
 *     for the HP-33120 function generator at ib-address 10
 * see further documentation (to be written!)
 *
 * compile with: gcc -c -O6 -m486 -D__KERNEL__ -DLINUX gpib_drv.c
 * or use Makefile!
 *
 * This is beta software !!!
 * use on Your own risk!
 *
 */

/* Kernel includes */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>

#include <linux/fs.h>
#include <linux/major.h>
#include <asm/uaccess.h>
#include <asm/page.h>
#include <asm/irq.h>
#include <linux/signal.h>
#include <linux/version.h>
#include <linux/timer.h>

#include "gpib_drv.h"

//#define DEBUG

static int GPIB_MAJOR = 30;		/* character special major number */

BOARDSTAT BoardStat;

static struct device device[32];

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
static DECLARE_MUTEX(gpib_wait_sem);
static DECLARE_MUTEX(gpib_busy_sem);
#else /* LINUX_24 */
static struct semaphore gpib_wait_sem = MUTEX;
static struct semaphore gpib_busy_sem = MUTEX;
#endif /* LINUX_24 */

static int ibsic(void);
static int ibonl(int);
void gpib_WaitForInt(char imr1mask, char imr2mask);

/*static volatile long busy = 0;*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
static DECLARE_WAIT_QUEUE_HEAD(gpib_wait);
#else /* LINUX_24 */
static struct wait_queue *gpib_wait = NULL;
#endif /* LINUX_24 */
static struct timer_list gpib_timer;
rwlock_t gpibhw_lock = RW_LOCK_UNLOCKED;
unsigned long flags;

void gpib_lock(void)
{
	down(&gpib_busy_sem);
}

void gpib_unlock(void)
{
	up(&gpib_busy_sem);
}

void gpib_timedout(unsigned long ptr)
{
	wake_up_interruptible(&gpib_wait);
}

static void gpib_Timeout(unsigned long ptr)
{
	up(&gpib_wait_sem);
	wake_up_interruptible(&gpib_wait);
	BoardStat.IBSTA |= TIMO;
}

static void gpib_StartTimer(long timeoutjif)
{
	init_timer(&gpib_timer);
	gpib_timer.function = gpib_Timeout;
	gpib_timer.data = 0;
	gpib_timer.expires = jiffies + timeoutjif;
	add_timer(&gpib_timer);
}

static void gpib_StopTimer()
{
	del_timer(&gpib_timer);
}

static void gpibsleep(long jif)
{
	init_timer(&gpib_timer);
	gpib_timer.function = gpib_timedout;
	gpib_timer.data = 0;
	gpib_timer.expires = jiffies + jif;
	add_timer(&gpib_timer);
	// printk("sleep on (%d, %d)\n", jiffies, jif);
	interruptible_sleep_on(&gpib_wait);
	// printk("sleep off (%d, %d)\n", gpib_timer.expires, jiffies);
}

void ibtimeout()
{
	int e1, e2;
	e1 = ibonl(1);
	e2 = ibsic();
}

/*
 * IBCMD
 * Write cnt command bytes from buf to the IEEE.
 */
static void ibcmd(char *buf, int cnt, int dev)
{
	const int jifdelta = 1;
	char s = 0;
	volatile int tries;
	volatile int jif = device[dev].timeout;
	volatile startjif, stopjif;

	gpib_StartTimer(200);

	if (In(adsr) & HR_NATN) {
		Out(auxmr, AUX_TCA);	/* if standby, go to CAC */
		if (In(isr1) & (HR_DEC | HR_DO | HR_DI))
			In(dir);
	}

	BoardStat.IBCNT = 0;
	BoardStat.IBSTA &= ~TIMO;

	while (BoardStat.IBCNT < cnt) {
		BoardStat.IntStat2 &= ~HR_CO;	/* Clear saved copy of Command Out */

		Out(cdor, buf[BoardStat.IBCNT++]);	/* Output command and increment count */

#ifdef USEINT
		gpib_WaitForInt(0, HR_CO);
#else
		while ((((BoardStat.IntStat2 |= In(isr2)) & HR_CO) == 0) && !(BoardStat.IBSTA & TIMO)) {
			schedule();
		}
#endif
		if (BoardStat.IBSTA & TIMO) {
			printk("gpib: cmd timeout\n");
			ibtimeout();
			return;
		}

		BoardStat.IBSTA = (s & HR_SRQI) ? SRQI : 0;
	}
	gpib_StopTimer();
}

static ibstat()
{								/* Update the GPIB status info */
	BoardStat.IBSTA &= ~SRQI;	/* Clear SRQI bit */
	BoardStat.IBSTA |= (In(isr2) & HR_SRQI) ? SRQI : 0;	/* Retest SRQI bit */
	return BoardStat.IBSTA;
}

/*
 * IBSRE
 * Send REN true if v is non-zero or false if v is zero.
 */
ibsre(v)
{
	Out(auxmr, (v ? AUX_SREN : AUX_CREN));	/* set or clear REN */
	return ibstat();
}

/*
 * IBONL
 * Initialize the interface hardware.  If v is non-zero then
 * the GPIB chip is enabled online.  If v is zero then it is
 * left disabled and offline.
 */
ibonl(v)
{
	Out(auxmr, AUX_CR);			/* reset chip */
	if (v)
		Out(auxmr, AUX_PON);	/* release pon state to bring online */

	(void) In(cptr);			/* clear registers by reading */
	(void) In(isr1);
	(void) In(isr2);

	Out(imr1, 0);				/* disable all interrupts */
	Out(imr2, 0);
	In(isr1);					/* read to clear */
	In(isr2);
	Out(spmr, 0);
	Out(adr, (PAD & LOMASK));	/* set GPIB address; MTA= 100, MLA= 040 */
#if (SAD)
	Out(adr, HR_ARS | (SAD & LOMASK));	/* enable secondary addressing */
	Out(admr, HR_TRM1 | HR_TRM0 | HR_ADM1);
#else
	Out(adr, HR_ARS | HR_DT | HR_DL);	/* disable secondary addressing */
	Out(admr, HR_TRM1 | HR_TRM0 | HR_ADM0);
#endif
	Out(eosr, 0);
	Out(auxmr, AUX_CIFC);		/* by default enable system controller */
	Out(auxmr, ICR | CLOCK);	/* set internal counter register */
	Out(auxmr, PPR | HR_PPU);	/* parallel poll unconfigure */
	Out(auxmr, AUXRA | 0);
//  Out(auxmr, AUXRB | HR_INV); /* set INT pin to active low */
	Out(auxmr, AUXRE | 0);

	if (!v) {
		ibcmd(UNT UNL, 2, 0);
		ibsre(0);
	}
	BoardStat.IBCNT = 0;
	BoardStat.HandshakeHeld = 0;	/* initialize handshake holdoff flag */
	return ibstat();
}

/*
 * IBSIC
 * Send IFC for at least 100 microseconds.
 */
static int ibsic(void)
{
	Out(auxmr, AUX_SIFC);		/* assert IFC */
	gpibsleep(1);
	Out(auxmr, AUX_CIFC);		/* clear IFC */
	BoardStat.IBSTA = (In(isr2) & HR_SRQI) ? SRQI : 0;
	return BoardStat.IBSTA;
}

static int gpib_status(char *buf, int count)
{
	static char version[64] = GPIB_VERSION;
	int ds = 32 * (sizeof(struct device));
	int out = count > (ds + 64) ? ds : count - 64;
	char *ap = (void *) device;
	int c;
	if (out < 64)
		return 0;

	for (c = 0; c < 64; ++c)
		put_user(version[c], buf++);

	c = out;
	while (c--)
		put_user(*ap++, buf++);

	return (out + 64);
}

/*
 * IBRD
 * Read up to cnt bytes of data from the GPIB into buf.  In
 * addition to I/O complete, the read operation terminates
 * on detection of EOI or timeout setting.
 */
static char *ibrd(char *buf, int cnt)
{
	char c;
#ifdef DEBUG
	int startjif = jiffies, start2jif, loops = 0;
#endif

	BoardStat.IBCNT = 0;

	Out(imr1, 0);				/* clear any previously arrived bits */
	Out(imr2, 0);

	BoardStat.IntStat1 &= ~(HR_DI | HR_END);

	BoardStat.IntStat2 = In(isr2);	/* clear the status registers */
	BoardStat.IntStat1 = In(isr1) | In(isr1);
	/* read isr1 twice in case of END delay */

	Out(auxmr, AUX_GTS);		/* if CAC, go to standby */

	if (BoardStat.HandshakeHeld) {
		Out(auxmr, AUXRA | HR_HLDA);
		Out(auxmr, AUX_FH);		/* finish handshake, if held off */
		BoardStat.HandshakeHeld = 0;
		BoardStat.IntStat2 |= In(isr2);	/* clear the status registers */
		BoardStat.IntStat1 |= In(isr1);
	} else if ((BoardStat.IntStat1 & HR_DI) && (BoardStat.IntStat1 & HR_END)) {
		printk("one-byte read with END\n");
		Out(auxmr, AUXRA | HR_HLDA);
		c = In(dir);
		put_user(c, buf);
		BoardStat.IBCNT = 1;
		return;
	}
	// Out(eosr, device[dev].eos); /* set eos-byte */
	Out(auxmr, AUXRA | HR_HLDE);	/* set hold off on END */

	while (BoardStat.IBCNT < cnt) {

#ifdef USEINT
		gpib_WaitForInt(HR_DI | HR_END, 0);
#ifdef DEBUG
		loops++;
#endif
#else
		while (!((BoardStat.IntStat1 |= In(isr1)) & HR_DI) && !(BoardStat.IBSTA & TIMO)) {	/* wait 
																							 * for
																							 * data 
																							 * in */
			schedule();
#ifdef DEBUG
			loops++;
#endif
		}
#endif

#ifdef DEBUG
		if (BoardStat.IBCNT == 0)
			start2jif = jiffies;
#endif

		if (BoardStat.IBSTA & TIMO) {
			printk("gpib: read timeout\n");
			BoardStat.IBSTA |= (In(isr2) & HR_SRQI) ? SRQI : 0;
			BoardStat.HandshakeHeld = 1;
			ibtimeout();
			return;
		}

		if (BoardStat.IBCNT == (cnt - 1))
			Out(auxmr, AUXRA | HR_HLDA);	// set hold off on all

		c = In(dir);
		put_user(c, buf);		// store byte and increment count
		buf++;
		BoardStat.IBCNT++;
		BoardStat.IntStat1 &= ~(HR_DI);

		if ((BoardStat.IntStat1 & HR_END)) {
			BoardStat.IntStat1 &= ~(HR_END);
			BoardStat.IBSTA |= END;
			break;				/* stop reading if EOI detected */
		}
	}

	BoardStat.IBSTA |= (BoardStat.IntStat2 & HR_SRQI) ? SRQI : 0;
	// Out(auxmr, AUXRA | HR_HLDA);
	// Out(auxmr, AUX_FH); /* finish handshake, if held off */
	BoardStat.HandshakeHeld = 1;
#ifdef DEBUG
	if (jiffies - start2jif == 0) {
		printk("gpib: read %d bytes in %d/%d jiffies with %d loops\n", BoardStat.IBCNT,
			   start2jif - startjif, jiffies - start2jif, loops);
	} else {
		printk("gpib: read %d bytes in %d/%d jiffies with %d loops (%d Bytes/s)\n", BoardStat.IBCNT,
			   start2jif - startjif, jiffies - start2jif, loops,
			   BoardStat.IBCNT * 100 / (jiffies - start2jif));
	}
#endif
	return buf;
}

static ssize_t gpib_read(struct file *file, char *fbuffer, size_t count, loff_t * dummy)
{
	int dev, cnt;
	static char cmd[4] = UNL UNT " ";
	char *buf;

	gpib_lock();

	dev = MINOR(file->f_dentry->d_inode->i_rdev);

	if (!dev) {
		gpib_unlock();
		return gpib_status(fbuffer, count);
	}

	if (count <= 0 || dev < 1 || dev > 31) {
		gpib_unlock();
		return -EINVAL;
	}

	ibsre(1);
	cmd[3] = 0x40 + dev;
	ibcmd(cmd, 4, dev);
/*     Out(auxmr,0x13); */

	gpib_StartTimer(device[dev].timeout);
	buf = ibrd(fbuffer, count);
	gpib_StopTimer();

/*     Out(auxmr,0x1c); */

	device[dev].status = BoardStat.IBSTA;

	gpib_unlock();
	return BoardStat.IBCNT;
}

/*
 * IBWRT
 * Write cnt bytes of data from buf to the GPIB.  The write
 * operation terminates only on I/O complete.  By default,
 * EOI is always sent along with the last byte.
 * Timeout setting terminates write operation.
 */

static void ibwrt(const char *buf, int cnt, int dev)
{
	int c, realcnt = cnt;
#ifdef DEBUG
	int startjif = jiffies, start2jif, loops = 0;
#endif

	Out(auxmr, AUX_GTS);		/* if active controller, go to standby */
	BoardStat.IBCNT = 0;

	if (device[dev].sendlastbyte)
		realcnt++;				/* to send one more byte */
	if (device[dev].crlf & 2)
		realcnt += 2;			/* new: send full cr/lf */

	while (BoardStat.IBCNT < (realcnt - 1)) {
		BoardStat.IntStat1 &= ~HR_DO;	/* clear saved copy of Data Out */

		if (BoardStat.IBCNT >= cnt + 1)
			c = 10;
		else if (BoardStat.IBCNT >= cnt)
			c = 13;
		else
			get_user(c, buf++);

		Out(cdor, c);			/* output data and increment count */
		BoardStat.IBCNT++;

#ifdef USEINT
		gpib_WaitForInt(HR_DO, 0);
#ifdef DEBUG
		loops++;
#endif
#else
		while (!((BoardStat.IntStat1 |= In(isr1)) & HR_DO) && !(BoardStat.IBSTA & TIMO)) {	/* wait 
																							 * for
																							 * Data 
																							 * Out; 
																							 * "or" 
																							 * to
																							 * preserve 
																							 * SRQI 
																							 */
			schedule();
#ifdef DEBUG
			loops++;
#endif
		}
#endif

#ifdef DEBUG
		if (BoardStat.IBCNT == 1)
			start2jif = jiffies;
#endif

		if (BoardStat.IBSTA & TIMO) {
			printk("gpib: write timeout\n");
			ibtimeout();
			return;
		}
	}
	Out(auxmr, AUX_SEOI);		/* last byte to send: Send EOI */

	if (BoardStat.IBCNT >= cnt)
		c = 10;
	else
		get_user(c, buf++);

	Out(cdor, c);

	BoardStat.IBCNT++;

#ifdef USEINT
	gpib_WaitForInt(HR_DO, 0);
#ifdef DEBUG
	loops++;
#endif
#else
	while (!((BoardStat.IntStat1 |= In(isr1)) & HR_DO) && !(BoardStat.IBSTA & TIMO)) {	/* wait for 
																						 * Data
																						 * Out;
																						 * "or" to
																						 * preserve 
																						 * SRQI */
		schedule();
#ifdef DEBUG
		loops++;
#endif
	}
#endif

	if (BoardStat.IBSTA & TIMO) {
		printk("gpib: write lastbyte timeout\n");
		ibtimeout();
		return;
	}
#ifdef DEBUG
	if (start2jif - startjif == 0) {
		printk("gpib: wrote %d bytes in %d,%d jiffies with %d loops\n", BoardStat.IBCNT,
			   start2jif - startjif, jiffies - start2jif, loops);
	} else {
		printk("gpib: wrote %d bytes in %d,%d jiffies with %d loops (%d Bytes/s)\n",
			   BoardStat.IBCNT, start2jif - startjif, jiffies - start2jif, loops,
			   BoardStat.IBCNT * 100 / (start2jif - startjif));
	}
#endif
}

static ssize_t gpib_write(struct file *file, const char *fbuffer, size_t count, loff_t * dummy)
{
	int dev;
	static char cmd[4] = UNL UNT "@";
	int tmp;

	gpib_lock();

	dev = MINOR(file->f_dentry->d_inode->i_rdev);
	if (count <= 0 || dev < 1 || dev > 31) {
		gpib_unlock();
		return -EINVAL;
	}

	/* checkbusy(dev); ++busy; */

	ibsre(1);
	cmd[3] = 0x20 + dev;
	ibcmd(cmd, 4, dev);
/*     Out(auxmr,0x1c); */
/*     ibcmd (&cmd[3],1,dev); */

	gpib_StartTimer(device[dev].timeout);
	ibwrt(fbuffer, count, dev);
	gpib_StopTimer();
	tmp = BoardStat.IBCNT;
/*    ibcmd (UNT UNL,2,dev); */

	device[dev].status = BoardStat.IBSTA;
	/* --busy; */
	gpib_unlock();
	return BoardStat.IBCNT = tmp;
}

static int gpib_open(struct inode *inode, struct file *filep)
{
	int dev;
	dev = MINOR(inode->i_rdev);
	if (device[dev].opened)
		return -EBUSY;
	device[dev].opened = 1;
	MOD_INC_USE_COUNT;
	return 0;
}

static int gpib_close(struct inode *inode, struct file *filep)
{
	int dev;
	dev = MINOR(inode->i_rdev);
	if (device[dev].opened == 0)
		return;
	device[dev].opened = 0;
	MOD_DEC_USE_COUNT;
	return 0;
}

static int gpib_ioctl(struct inode *inode, struct file *filep, unsigned int cmd, unsigned long arg)
{
	int dev;
	static char cmdstr[4];
	int i = 0;
	int rc = dev;

	gpib_lock();

	dev = MINOR(inode->i_rdev);

	if (dev < 0 || dev > 31) {
		gpib_unlock();
		return -EINVAL;
	}

	switch (cmd) {
	case GPIB_NUL:
		gpib_unlock();
		return dev;
		break;
	case GPIB_RST:
		ibonl(1);
		ibsic();
		gpib_unlock();
		return dev;
		break;
	}

	switch (cmd) {
		/* adressed IEEE-Commands */
		/* work here only on the Device with Adress dev! */
	case GPIB_GTL:
		cmdstr[0] = '?';
		cmdstr[1] = 0x20 + dev;
		cmdstr[2] = 0x01;
		ibcmd(cmdstr, 3, dev);
		break;
	case GPIB_SDC:
		cmdstr[0] = '?';
		cmdstr[1] = 0x20 + dev;
		cmdstr[2] = 0x04;
		ibcmd(cmdstr, 3, dev);
		break;
/*         case GPIB_PPC: */
/*             break; */
	case GPIB_GET:
		cmdstr[0] = '?';
		cmdstr[1] = 0x20 + dev;
		cmdstr[2] = 0x08;
		ibcmd(cmdstr, 3, dev);
		break;
/*         case GPIB_TCT:  */
/*             break; */

		/* universal IEEE-Commands */
	case GPIB_LLO:
		ibcmd("\0x11", 1, dev);
		break;
	case GPIB_DCL:
		ibcmd("\0x14", 1, dev);
		break;
/*         case GPIB_PPU: */
/*             break; */
/*         case GPIB_SPE: */
/*             break; */
/*         case GPIB_SPD: */
/*             break; */

		/* Management Line Commands */
	case GPIB_IFC:
		ibsic();
		break;
	case GPIB_SREN:
		ibsre(1);
		cmdstr[0] = '_';
		cmdstr[1] = '?';
		cmdstr[2] = 0x20 + dev;
		ibcmd(cmdstr, 3, dev);
		break;
	case GPIB_CREN:
		ibsre(0);
		break;

		/* Driver Commands */
	case GPIB_ONL:
		ibonl(1);
		break;
/*         case GPIB_CMD: */
/*      break; */

		/* Device Setting Commands */
	case GPIB_DEV_EOF:
		device[dev].sendeof = arg;
		device[dev].eof = 0;
		break;
	case GPIB_DEV_CRLF:
		device[dev].crlf = arg;
		device[dev].eof = 0;
		break;
	case GPIB_DEV_TIMEOUT:
		device[dev].timeout = arg;
		break;
	case GPIB_DEV_SETEOS:
		device[dev].eos = arg;
		break;
	case GPIB_DEV_ENDONEOS:
		device[dev].endoneos = arg;
		break;
	case GPIB_DEV_SHAREDEV:
		device[dev].shared = 1;
		device[dev].opened = 0;
		break;
	case GPIB_DEV_SENDLASTBYTE:
		device[dev].sendlastbyte = arg;
		device[dev].eof = 0;
		break;
		/* Device Query Commands */
	case GPIB_ERR:
		device[dev].status = 0;
		rc = device[dev].error = 0;

		break;
	case GPIB_STA:
		if (dev) {
			rc = device[dev].status;
			device[dev].status = 0;
		} else
			rc = ibstat();
		break;
	default:
		/*--busy;*/
		gpib_unlock();
		return -ENOSYS;
	}

/*    ibcmd (UNT UNL,2,dev); */

	if (dev) {
		device[dev].status = BoardStat.IBSTA;
	}
	gpib_unlock();
	/* --busy; */
	return rc;
}

static struct file_operations gpib_fops;

/*
 * And now the modules code and kernel interface.
 */

extern int printk(const char *fmt, ...);

static char *gpib = "gpib";

#ifdef USEINT
void gpib_WaitForInt(char imr1mask, char imr2mask)
{
	down(&gpib_wait_sem);

	outb_p(0, PC2A_CLR_IRQ_ADR + IRQ);
	Out(imr1, imr1mask);
	/* now it's time to enable board interrupts */
	Out(imr2, imr2mask);

	/* now push process to sleep */
	down(&gpib_wait_sem);
	up(&gpib_wait_sem);
}

static void gpib_interrupt(int irq, void *devId, struct pt_regs *regs)
{
	char c, GPIBCommand[4];

	BoardStat.IntStat1 |= In(isr1);
	BoardStat.IntStat2 |= In(isr2);

	Out(imr1, 0);
	Out(imr2, 0);
	up(&gpib_wait_sem);
}
#endif

int init_module(void)
{
	int i, e1, e2;

	gpib_lock();

	e1 = ibonl(1);
	e2 = ibsic();

	printk("gpib_drv: init_module called (%d,%d)\n", e1, e2);

	gpib_fops.llseek = NULL;
	gpib_fops.read = gpib_read;
	gpib_fops.write = gpib_write;
	gpib_fops.readdir = NULL;
	gpib_fops.poll = NULL;
	gpib_fops.ioctl = gpib_ioctl;
	gpib_fops.mmap = NULL;
	gpib_fops.open = gpib_open;
	gpib_fops.flush = NULL;
	gpib_fops.release = gpib_close;
	gpib_fops.fsync = NULL;
	gpib_fops.fasync = NULL;
	gpib_fops.lock = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	gpib_fops.readv = NULL;
	gpib_fops.writev = NULL;
	gpib_fops.sendpage = NULL;
	gpib_fops.get_unmapped_area = NULL;
#else /* LINUX_24 */
	gpib_fops.check_media_change = NULL;
	gpib_fops.revalidate = NULL;
#endif /* LINUX_24 */

	if (register_chrdev(GPIB_MAJOR, gpib, &gpib_fops)) {
		printk("register_chrdev failed\n");
		return -EIO;
	}
#if USEINT
	if (request_irq(IRQ, gpib_interrupt, SA_INTERRUPT | SA_SAMPLE_RANDOM, gpib, NULL)) {
		printk("request_irq failed\n");
		return -EIO;
	}
#endif

	for (i = 0; i < 32; ++i) {
		device[i].opened = 0;
		device[i].eof = 0;
		device[i].shared = 0;
		device[i].sendeof = 0;
		device[i].crlf = 0;
		device[i].timeout = TOJIF;
		device[i].eos = LF;
		device[i].endoneos = 0;
		device[i].sendlastbyte = 0;
		device[i].status = 0;
	}

	BoardStat.IntStat1 = 0;
	BoardStat.IntStat2 = 0;
	BoardStat.HandshakeHeld = 0;
	gpib_unlock();

	return 0;
}

void cleanup_module(void)
{
	printk("gpib_drv: cleanup_module called\n");

	gpib_lock();

	ibsic();
	// ibonl(0);

	if (MOD_IN_USE)
		printk("gpib: device busy, remove delayed\n");

#ifdef USEINT
	free_irq(IRQ, NULL);
#endif
	if (unregister_chrdev(GPIB_MAJOR, "gpib") != 0) {
		printk("cleanup_module failed\n");
	} else {
		printk("cleanup_module succeeded\n");
	}

	gpib_unlock();
}
