/*
 * General purpuse IEEE character device driver
 * for NEC 7210 based cards (and compatible).
 * Tested with 7210 and NI PC-IIa
 *
 * Written & (C) 1995,96 by Michael Boegl und Thomas Hohenberger
 * Lehrstuhl Experimentalphysik V
 * Universitaet Bayreuth
 * e-mail: ieee@lowt.phy.uni-bayreuth.de
 * 
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation with the modifications stated in the 
 *   'COPYING' licence file.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   (See file 'GPL' distributed with this package.)
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This is beta software !!!
 * use on Your own risk!
 *
 */

#ifndef _GPIB_H_
#define _GPIB_H_

#ifndef GPIB_VERSION
#define GPIB_VERSION "V1.0beta"
#endif

#define SLEEPCOUNT 100000L

#ifdef __KERNEL__

/* Interface board access macros */

#ifdef GPIBIO
#define In(x)           inp(IB + x)
#define Out(x,a)        outp(a, (IB + x))
#else
#include <asm/io.h>
#define In(x)           inb_p(IB + x)
#define Out(x,a)        outb_p(a, (IB + x))
#endif

/* Editable parameters (use Makefile configuration if possible !) */

#define PAD        0			/* Primary address of GPIB   */
#define SAD        0			/* Secondary address of GPIB */
								/* 60H - 7EH, or 0 if none   */

#ifndef PC2A
#define PC2A       1			/* board selection flag      */
								/* 1 if GPIB-PC2A, 0 if GPIB-PC2 */
#endif

#ifndef IB
#if(PC2A==1)
#define IB      0x2E1			/* I/O port address of interface board */
#else /* ex:  0x2E1 for GPIB-PC2A  */
#define IB      0x2B8			/* 0x2B8 for GPIB-PC2   */
#endif
#endif

#ifndef IRQ
#define IRQ        7			/* IRQ used by gpib */
#endif

#ifndef DMA
#define CH         1			/* DMA channel used by GPIB-PC */
#else
#define CH         DMA
#endif

#define CLOCK      8			/* Clock of NEC7210 (5/8 ?) */

#define TOJIF      300			/* Timeout in Jiffies 0.5 s */
#define            LF 0x0a		/* Standard Abschlusszeichen bei gpib_read */

#endif	/* __KERNEL__ */

/* nothing should be changed below */

/* Commands to be sent to gpib_drv via ioctl(): */
   /* Addressed Command Group */
#define GPIB_GTL           0x01	/* GO TO LOCAL */
#define GPIB_SDC           0x04	/* SELECTIVE DEVICE CLEAR */
#define GPIB_PPC           0x05	/* PARALLEL POLL CONFIGURE */
#define GPIB_GET           0x08	/* GROUP EXECUTIVE TRIGGER */
#define GPIB_TCT           0x09	/* TRANSFER CONTROL TO */
   /* Universal Command Group */
#define GPIB_LLO           0x11	/* LOCAL LOCKOUT */
#define GPIB_DCL           0x14	/* DEVICE CLEAR */
#define GPIB_PPU           0x15	/* PARALLEL POLL UNCONFIGURE */
#define GPIB_SPE           0x18	/* SERIAL POLL ENABLE */
#define GPIB_SPD           0x19	/* SERIAL POLL DISABLE */
   /* Management Line Commands */
#define GPIB_IFC           0x30	/* INTERFACE CLEAR */
#define GPIB_SREN          0x31	/* SET REMOTE ENABLE LINE */
#define GPIB_CREN          0x32	/* CLEAR REMOTE ENABLE LINE */
   /* Driver Commands */
#define GPIB_NUL           0x20	/* GIVE NULL COMMAND TO DRIVER */
#define GPIB_ONL           0x21	/* SET DRIVER ONLINE */
#define GPIB_RST           0x22	/* GPIB RESET (ALSO IF DEVICE BUSY!) */
#define GPIB_CMD           0x23	/* SENDS COMMAND STRING TO IEEE */
   /* Device Settings Commands */
#define GPIB_DEV_EOF       0x41
#define GPIB_DEV_CRLF      0x42
#define GPIB_DEV_TIMEOUT   0x43
#define GPIB_DEV_SETEOS    0x44
#define GPIB_DEV_SHAREDEV  0x45
#ifdef SLOW
#define GPIB_DEV_SETDELAY  0x46
#endif
#define GPIB_DEV_ENDONEOS  0x47
#define GPIB_DEV_SENDLASTBYTE 0x48
   /* Device Driver Query Commands */
#define GPIB_ERR           0x50
#define GPIB_STA           0x51

#define UNL "?"
#define UNT "_"

/* C user command interface */

#define GPIB_PATH "/dev/gpib/"
#define GPIB_ETCPATH "/etc/gpib/"

#define ib_sic(dev)           ioctl(dev,GPIB_IFC)
#define ib_ifc(dev)           ib_sic(dev)
#define ib_open(dev)          open(GPIB_PATH dev,O_RDWR|O_SYNC)
#define ib_close(dev)         close(dev)
#define ib_read(dev,str,len)  read(dev,str,len)
#define ib_write(dev,str,len) write(dev,str,len)
#define ib_ctl                ioctl
#define ib_ioctl              ioctl
#define ib_error(dev)         ioctl(dev,GPIB_ERR)
#define ib_status(dev)        ioctl(dev,GPIB_STA)

struct device
{
	char opened;
	char shared;
	char sendeof;
	char eof;
	char crlf;
	char eos;
	char endoneos;
	char sendlastbyte;
	short timeout;
	short error;
	short status;
#ifdef SLOW
	short delay;
	char sleeps;
	char pad;
#endif
};

#ifdef __KERNEL__

static ibstat();

/* Status bits, returned by all functions */

#define TIMO    (1<<14)			/* TIMEOUT occurred before call was completed 
								 */
#define END     (1<<13)			/* EOI terminated the ibrd call */
#define SRQI    (1<<12)			/* SRQ was asserted before or during the call 
								 */

/* Hardware register definitions */

/* Write-only registers          */

#define PC2A_CLR_IRQ_ADR    0x02f0

#if (PC2A)

#define cdor    0				/* +0 byte out register                */
#define imr1    0x400			/* +1 interrupt mask register 1        */
#define imr2    0x800			/* +2 interrupt mask register 2        */
#define spmr    0xC00			/* +3 serial poll mode register        */
#define admr    0x1000			/* +4 address mode register            */
#define auxmr   0x1400			/* +5 auxiliary mode register          */
#define adr     0x1800			/* +6 address register 0/1             */
#define eosr    0x1C00			/* +7 end of string register           */

#else

#define cdor    0				/* +0 byte out register                */
#define imr1    1				/* +1 interrupt mask register 1        */
#define imr2    2				/* +2 interrupt mask register 2        */
#define spmr    3				/* +3 serial poll mode register        */
#define admr    4				/* +4 address mode register            */
#define auxmr   5				/* +5 auxiliary mode register          */
#define adr     6				/* +6 address register 0/1             */
#define eosr    7				/* +7 end of string register           */

#endif

/* Read-only register mnemonics corresponding to write-only registers */

#define dir     cdor			/* +0 data in register                  */
#define isr1    imr1			/* +1 interrupt status register 1       */
#define isr2    imr2			/* +2 interrupt status register 2       */
#define spsr    spmr			/* +3 serial poll status register       */
#define adsr    admr			/* +4 adress status register            */
#define cptr    auxmr			/* +5 command pass through              */
#define adr0    adr				/* +6 adress 0                          */
#define adr1    eosr			/* +7 adress 1                          */

/* Control masks for hidden registers (auxmr) */

#define ICR     0040
#define PPR     0140
#define AUXRA   0200
#define AUXRB   0240
#define AUXRE   0300

#define LOMASK  0x1F			/* mask to specify lower 5 bits */

/* Hardware register bit definitions */
/*      Name            Bit(s)          register        */
#define HR_DI           (1<<0)	/* isr1         */
#define HR_DO           (1<<1)	/* isr1         */
#define HR_DEC          (1<<3)	/* isr1         */
#define HR_END          (1<<4)	/* isr1         */
#define HR_CO           (1<<3)	/* isr2         */
#define HR_SRQI         (1<<6)	/* isr2         */
#define HR_DMAI         (1<<4)	/* imr2         */
#define HR_DMAO         (1<<5)	/* imr2         */
#define HR_NATN         (1<<6)	/* adsr         */
#define HR_ADM0         (1<<0)	/* admr         */
#define HR_ADM1         (1<<1)	/* admr         */
#define HR_TRM0         (1<<4)	/* admr         */
#define HR_TRM1         (1<<5)	/* admr         */
#define HR_DL           (1<<5)	/* adr          */
#define HR_DT           (1<<6)	/* adr          */
#define HR_ARS          (1<<7)	/* adr          */
#define HR_INV          (1<<3)	/* auxrb        */
#define HR_PPU          (1<<4)	/* ppr          */
#define HR_HLDE         (1<<1)	/* auxra        */
#define HR_HLDA         (1<<0)	/* auxra        */

/* 7210 Auxiliary Commands */

#define AUX_PON         000		/* Immediate Execute pon        */
#define AUX_CR          002		/* Chip Reset                   */
#define AUX_FH          003		/* Finish Handshake             */
#define AUX_SEOI        006		/* Send EOI                     */
#define AUX_TCA         021		/* Take Control Asynchronously  */
#define AUX_GTS         020		/* Go To Standby                */
#define AUX_EPP         035		/* Execute Parallel Poll        */
#define AUX_SIFC        036		/* Set IFC                      */
#define AUX_CIFC        026		/* Clear IFC                    */
#define AUX_SREN        037		/* Set REN                      */
#define AUX_CREN        027		/* Clear REN                    */

#endif	/* __KERNEL__ */

/**************************************************************************
  *
  *                 Board Status Structure
  *
  *  This structure contains information that describes the current
  *  state of a GPIB board.
  *
   *************************************************************************/

struct BS
{
	long ByteCount;				/* Number of bytes to transfer */
	char *BufPtr;				/* Pointer to current data buffer */
	char EndByte;				/* Final byte of write */
	int SerialPollByte;			/* Current serial poll response byte */
	int IntStat1;				/* Current value of 7210's ISR Stat 1 reg */
	int IntStat2;				/* Current value of 7210's ISR Stat 2 reg */
	int CurAUXRAVal;			/* Current value of 7210 AUX A register */
	short IsBoard;
	short IBSTA;				/* Library status byte */
	short IBERR;				/* library error code */
	long IBCNT;					/* Current count */
	void (*SRQHandler) (void);	/* Ptr to SRQ interrupt handler */
	int Timeout;				/* Timeout value for board */
	int HandshakeHeld;			/* TRUE=Handshake * is * * * currently * * *
								 * held * * * off int ExtraByteFlag; *
								 * TRUE=Extra * * * byte * * * is * * *
								 * waiting * * * */
	char ExtraByte; /* Extra * byte's * value * * */
	int IntCount; /* Count * of * * interrupts * (for * * debugging) */
	int IntLevel; 
	int EOTMode; /* Termination * * for * write * */
};
typedef struct BS BOARDSTAT;

#endif	/* _GPIB_H_ */
